@@ -1,31 +1,32 @@
 """Classes used to build a schema.
 
 :organization: Logilab
-:copyright: 2003-2008 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
+:copyright: 2003-2009 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
 :contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
 :license: General Public License version 2 - http://www.gnu.org/licenses
 """
 __docformat__ = "restructuredtext en"
 
+from warnings import warn
+
 from logilab.common import attrdict
 from logilab.common.compat import sorted
+from logilab.common.decorators import iclassmethod
 
 from yams import BASE_TYPES, MARKER, BadSchemaDefinition
 from yams.constraints import (SizeConstraint, UniqueConstraint,
-                              StaticVocabularyConstraint)
+                              StaticVocabularyConstraint, format_constraint)
 
 __all__ = ('ObjectRelation', 'SubjectRelation', 'BothWayRelation',
-           'RelationDefinition', 'EntityType', 'MetaEntityType',
-           'RestrictedEntityType', 'UserEntityType', 'MetaUserEntityType',
-           'RelationType', 'MetaRelationType', 'UserRelationType',
-           'MetaUserRelationType', 'AttributeRelationType',
-           'MetaAttributeRelationType',
+           'RelationDefinition', 'EntityType', 'RelationType',
            'SubjectRelation', 'ObjectRelation', 'BothWayRelation',
-           ) + tuple(BASE_TYPES)
+           'RichString', ) + tuple(BASE_TYPES)
 
-ETYPE_PROPERTIES = ('meta', 'description', 'permissions')
-# don't put description inside, handled "manually"
-RTYPE_PROPERTIES = ('meta', 'symetric', 'inlined', 'fulltext_container', 'permissions')
+ETYPE_PROPERTIES = ('description', 'permissions',
+                    'meta') # XXX meta is deprecated
+# don't put description inside, handled "manualy"
+RTYPE_PROPERTIES = ('symetric', 'inlined', 'fulltext_container', 'permissions',
+                    'meta') # XXX meta is deprecated
 RDEF_PROPERTIES = ('cardinality', 'constraints', 'composite',
                    'order',  'default', 'uid', 'indexed', 'uid',
                    'fulltextindexed', 'internationalizable')
@@ -44,6 +45,12 @@ def _add_constraint(kwargs, constraint):
 
 def _add_relation(relations, rdef, name=None, insertidx=None):
     """Add relation (param rdef) to list of relations (param relations)."""
+    if isinstance(rdef, RichString):
+        format_attrdef = String(internationalizable=True,
+                                default=rdef.default_format, maxsize=50,
+                                constraints=rdef.format_constraints)
+        _add_relation(relations, format_attrdef,
+                      (name or rdef.name) + '_format', insertidx)
     if isinstance(rdef, BothWayRelation):
         _add_relation(relations, rdef.subjectrel, name, insertidx)
         _add_relation(relations, rdef.objectrel, name, insertidx)
@@ -59,7 +66,8 @@ def _check_kwargs(kwargs, attributes):
     """Check that all keys of kwargs are actual attributes."""
     for key in kwargs:
         if not key in attributes:
-            raise BadSchemaDefinition('no such property %r in %r' % (key, attributes))
+            raise BadSchemaDefinition('no such property %r in %r'
+                                      % (key, attributes))
 
 def _copy_attributes(fromobj, toobj, attributes):
     for attr in attributes:
@@ -68,13 +76,14 @@ def _copy_attributes(fromobj, toobj, attributes):
             continue
         ovalue = getattr(toobj, attr, MARKER)
         if not ovalue is MARKER and value != ovalue:
-            raise BadSchemaDefinition('conflicting values %r/%r for property %s of %s'
-                                      % (ovalue, value, attr, toobj))
+            raise BadSchemaDefinition(
+                'conflicting values %r/%r for property %s of %s'
+                % (ovalue, value, attr, toobj))
         setattr(toobj, attr, value)
 
 def register_base_types(schema):
     for etype in BASE_TYPES:
-        edef = EntityType(name=etype, meta=True)
+        edef = EntityType(name=etype)
         schema.add_entity_type(edef).set_default_groups()
 
 
@@ -100,13 +109,15 @@ class Definition(object):
     def __repr__(self):
         return '<%s %r @%x>' % (self.__class__.__name__, self.name, id(self))
 
-    def expand_type_definitions(self, defined):
+    @classmethod
+    def expand_type_definitions(cls, defined):
         """Schema building step 1: register definition objects by adding them
         to the `defined` dictionnary.
         """
         raise NotImplementedError()
 
-    def expand_relation_definitions(self, defined, schema):
+    @classmethod
+    def expand_relation_definitions(cls, defined, schema):
         """Schema building step 2: register all relations definition,
         expanding wildcard if necessary.
         """
@@ -143,6 +154,7 @@ class metadefinition(type):
         return defclass
 
 
+
 class EntityType(Definition):
     # FIXME reader magic forbids to define a docstring...
     #"""an entity has attributes and can be linked to other entities by
@@ -165,50 +177,55 @@ class EntityType(Definition):
         _check_kwargs(kwargs, ETYPE_PROPERTIES)
         _copy_attributes(attrdict(kwargs), self, ETYPE_PROPERTIES)
         # if not hasattr(self, 'relations'):
-        self.relations = list(self.__relations__)
+        #self.relations = list(self.__relations__)
         self.specialized_type = self.__class__.__dict__.get('__specializes__')
 
+    def __str__(self):
+        return 'entity type %r' % self.name
+
     @property
     def specialized_by(self):
         return self.__class__.__dict__.get('__specialized_by__', [])
 
-    def __str__(self):
-        return 'entity type %r' % self.name
-
-    def expand_type_definitions(self, defined):
+    @classmethod
+    def expand_type_definitions(cls, defined):
         """Schema building step 1: register definition objects by adding
         them to the `defined` dictionnary.
         """
-        assert self.name not in defined, "type '%s' was already defined" % self.name
-        self._defined = defined # XXX may be used later (eg .add_relation())
-        defined[self.name] = self
-        for relation in self.relations:
-            self._ensure_relation_type(relation)
+        name = getattr(cls, 'name', cls.__name__)
+        assert name not in defined, "type '%s' was already defined" % name
+        cls._defined = defined # XXX may be used later (eg .add_relation())
+        defined[name] = cls
+        for relation in cls.__relations__:
+            cls._ensure_relation_type(relation)
 
-    def _ensure_relation_type(self, relation):
+    @classmethod
+    def _ensure_relation_type(cls, relation):
         rtype = RelationType(relation.name)
         _copy_attributes(relation, rtype, RTYPE_PROPERTIES)
-        defined = self._defined
+        defined = cls._defined
         if relation.name in defined:
             _copy_attributes(rtype, defined[relation.name], RTYPE_PROPERTIES)
         else:
             defined[relation.name] = rtype
 
-    def expand_relation_definitions(self, defined, schema):
+    @classmethod
+    def expand_relation_definitions(cls, defined, schema):
         """schema building step 2:
 
         register all relations definition, expanding wildcards if necessary
         """
         order = 1
-        for relation in self.relations:
+        name = getattr(cls, 'name', cls.__name__)
+        for relation in cls.__relations__:
             if isinstance(relation, SubjectRelation):
-                rdef = RelationDefinition(subject=self.name, name=relation.name,
+                rdef = RelationDefinition(subject=name, name=relation.name,
                                           object=relation.etype, order=order)
                 _copy_attributes(relation, rdef, RDEF_PROPERTIES + ('description',))
             elif isinstance(relation, ObjectRelation):
                 rdef = RelationDefinition(subject=relation.etype,
                                           name=relation.name,
-                                          object=self.name, order=order)
+                                          object=name, order=order)
                 _copy_attributes(relation, rdef, RDEF_PROPERTIES + ('description',))
             else:
                 raise BadSchemaDefinition('dunno how to handle %s' % relation)
@@ -217,41 +234,51 @@ class EntityType(Definition):
 
     # methods that can be used to extend an existant schema definition ########
 
-    def extend(self, othermetadefcls):
+    @classmethod
+    def extend(cls, othermetadefcls):
         for rdef in othermetadefcls.__relations__:
-            self.add_relation(rdef)
+            cls.add_relation(rdef)
 
-    def add_relation(self, rdef, name=None):
+    @classmethod
+    def add_relation(cls, rdef, name=None):
         if name:
             rdef.name = name
-        self._ensure_relation_type(rdef)
-        _add_relation(self.relations, rdef, name)
+        cls._ensure_relation_type(rdef)
+        _add_relation(cls.__relations__, rdef, name)
+        if isinstance(rdef, RichString) and not rdef in cls._defined:
+            format_attr_name = (name or rdef.name) + '_format'
+            rdef = cls.get_relations(format_attr_name).next()
+            cls._ensure_relation_type(rdef)
 
-    def insert_relation_after(self, afterrelname, name, rdef):
+    @classmethod
+    def insert_relation_after(cls, afterrelname, name, rdef):
         # FIXME change order of arguments to rdef, name, afterrelname ?
         rdef.name = name
-        self._ensure_relation_type(rdef)
-        for i, rel in enumerate(self.relations):
+        cls._ensure_relation_type(rdef)
+        for i, rel in enumerate(cls.__relations__):
             if rel.name == afterrelname:
                 break
         else:
             raise BadSchemaDefinition("can't find %s relation on %s" % (
-                    afterrelname, self))
-        _add_relation(self.relations, rdef, name, i+1)
+                    afterrelname, cls))
+        _add_relation(cls.__relations__, rdef, name, i+1)
 
-    def remove_relation(self, name):
-        for rdef in self.get_relations(name):
-            self.relations.remove(rdef)
-            self.__relations__.remove(rdef)
+    @classmethod
+    def remove_relation(cls, name):
+        for rdef in cls.get_relations(name):
+            cls.__relations__.remove(rdef)
+            cls.__relations__.remove(rdef)
 
-    def get_relations(self, name):
+    @classmethod
+    def get_relations(cls, name):
         """get relation definitions by name (may have multiple definitions with
         the same name if the relation is both a subject and object relation)
         """
-        for rdef in self.relations[:]:
+        for rdef in cls.__relations__[:]:
             if rdef.name == name:
                 yield rdef
 
+
 class RelationType(Definition):
     symetric = MARKER
     inlined = MARKER
@@ -260,32 +287,40 @@ class RelationType(Definition):
     def __init__(self, name=None, **kwargs):
         """kwargs must have values in RTYPE_PROPERTIES"""
         super(RelationType, self).__init__(name)
+        if kwargs.pop('meta', None):
+            warn('meta is deprecated', DeprecationWarning, stacklevel=2)
         _check_kwargs(kwargs, RTYPE_PROPERTIES + ('description',))
         _copy_attributes(attrdict(kwargs), self, RTYPE_PROPERTIES + ('description',))
 
     def __str__(self):
         return 'relation type %r' % self.name
 
-    def expand_type_definitions(self, defined):
+    @classmethod
+    def expand_type_definitions(cls, defined):
         """schema building step 1:
 
         register definition objects by adding them to the `defined` dictionnary
         """
-        if self.name in defined:
-            _copy_attributes(self, defined[self.name],
+        name = getattr(cls, 'name', cls.__name__)
+        if cls.__doc__ and not cls.description:
+            cls.description = ' '.join(cls.__doc__.split())
+        if name in defined:
+            _copy_attributes(cls, defined[name],
                              REL_PROPERTIES + ('subject', 'object'))
         else:
-            defined[self.name] = self
+            defined[name] = cls
 
-    def expand_relation_definitions(self, defined, schema):
+    @classmethod
+    def expand_relation_definitions(cls, defined, schema):
         """schema building step 2:
 
         register all relations definition, expanding wildcard if necessary
         """
-        if getattr(self, 'subject', None) and getattr(self, 'object', None):
-            rdef = RelationDefinition(subject=self.subject, name=self.name,
-                                      object=self.object)
-            _copy_attributes(self, rdef, RDEF_PROPERTIES)
+        name = getattr(cls, 'name', cls.__name__)
+        if getattr(cls, 'subject', None) and getattr(cls, 'object', None):
+            rdef = RelationDefinition(subject=cls.subject, name=name,
+                                      object=cls.object)
+            _copy_attributes(cls, rdef, RDEF_PROPERTIES)
             rdef._add_relations(defined, schema)
 
 
@@ -314,6 +349,8 @@ class RelationDefinition(Definition):
         else:
             self.object = self.__class__.object
         super(RelationDefinition, self).__init__(name)
+        if kwargs.pop('meta', None):
+            warn('meta is deprecated', DeprecationWarning)
         _check_kwargs(kwargs, RDEF_PROPERTIES + ('description',))
         _copy_attributes(attrdict(kwargs), self, RDEF_PROPERTIES + ('description',))
         if self.constraints:
@@ -322,32 +359,36 @@ class RelationDefinition(Definition):
     def __str__(self):
         return 'relation definition (%(subject)s %(name)s %(object)s)' % self.__dict__
 
-    def expand_type_definitions(self, defined):
+    @classmethod
+    def expand_type_definitions(cls, defined):
         """schema building step 1:
 
         register definition objects by adding them to the `defined` dictionnary
         """
-        rtype = RelationType(self.name)
-        _copy_attributes(self, rtype, RTYPE_PROPERTIES)
-        if self.name in defined:
-            _copy_attributes(rtype, defined[self.name], RTYPE_PROPERTIES)
+        name = getattr(cls, 'name', cls.__name__)
+        rtype = RelationType(name)
+        _copy_attributes(cls, rtype, RTYPE_PROPERTIES)
+        if name in defined:
+            _copy_attributes(rtype, defined[name], RTYPE_PROPERTIES)
         else:
-            defined[self.name] = rtype
-        key = (self.subject, self.name, self.object)
+            defined[name] = rtype
+        key = (cls.subject, name, cls.object)
         if key in defined:
-            raise BadSchemaDefinition('duplicated %s' % self)
-        defined[key] = self
+            raise BadSchemaDefinition('duplicated %s' % cls)
+        defined[key] = cls
 
-    def expand_relation_definitions(self, defined, schema):
+    @classmethod
+    def expand_relation_definitions(cls, defined, schema):
         """schema building step 2:
 
         register all relations definition, expanding wildcard if necessary
         """
-        assert self.subject and self.object, '%s; check the schema' % self
-        self._add_relations(defined, schema)
+        assert cls.subject and cls.object, '%s; check the schema' % cls
+        cls()._add_relations(defined, schema)
 
     def _add_relations(self, defined, schema):
-        rtype = defined[self.name]
+        name = getattr(self, 'name', self.__class__.__name__)
+        rtype = defined[name]
         _copy_attributes(rtype, self, RDEF_PROPERTIES)
         # process default cardinality and constraints if not set yet
         cardinality = self.cardinality
@@ -362,91 +403,33 @@ class RelationDefinition(Definition):
             assert cardinality[1] in '1?+*'
         if not self.constraints:
             self.constraints = ()
-        rschema = schema.rschema(self.name)
-        for subj in self._actual_types(schema, self.subject):
-            for obj in self._actual_types(schema, self.object):
-                rdef = RelationDefinition(subj, self.name, obj)
+        rschema = schema.rschema(name)
+        if self.subject == '**' or self.object == '**':
+            warn('** is deprecated, use * (%s)' % rtype, DeprecationWarning)
+        for subj in _actual_types(schema, self.subject):
+            for obj in _actual_types(schema, self.object):
+                rdef = RelationDefinition(subj, name, obj)
                 _copy_attributes(self, rdef, RDEF_PROPERTIES + ('description',))
                 schema.add_relation_def(rdef)
 
-    def _actual_types(self, schema, etype):
-        if etype == '*':
-            return self._wildcard_etypes(schema)
-        elif etype == '**':
-            return self._pow_etypes(schema)
-        elif isinstance(etype, (tuple, list)):
-            return etype
-        return (etype,)
+def _actual_types(schema, etype):
+    if etype in ('**', '*'): # XXX ** is deprecated
+        return _pow_etypes(schema)
+    if isinstance(etype, (list, tuple)):
+        return etype
+    if not isinstance(etype, basestring):
+        raise RuntimeError('Entity types must not be instances but strings '
+                           'or list/tuples thereof. Ex. (bad, good) : '
+                           'SubjectRelation(Foo), SubjectRelation("Foo"). '
+                           'Hence, %r is not acceptable.' % etype)
+    return (etype,)
 
-    def _wildcard_etypes(self, schema):
-        for eschema in schema.entities():
-            if eschema.is_final() or eschema.meta:
-                continue
-            yield eschema.type
+def _pow_etypes(schema):
+    for eschema in schema.entities():
+        if eschema.is_final():
+            continue
+        yield eschema.type
 
-    def _pow_etypes(self, schema):
-        for eschema in schema.entities():
-            if eschema.is_final():
-                continue
-            yield eschema.type
-
-
-# various derivated classes with some predefined values #######################
-
-class MetaEntityType(EntityType):
-    permissions = {
-        'read':   ('managers', 'users', 'guests',),
-        'add':    ('managers',),
-        'delete': ('managers',),
-        'update': ('managers', 'owners',),
-        }
-    meta = True
-
-class RestrictedEntityType(MetaEntityType):
-    permissions = {
-        'read':   ('managers', 'users',),
-        'add':    ('managers',),
-        'delete': ('managers',),
-        'update': ('managers', 'owners',),
-        }
-
-class UserEntityType(EntityType):
-    permissions = {
-        'read':   ('managers', 'users', 'guests',),
-        'add':    ('managers', 'users',),
-        'delete': ('managers', 'owners',),
-        'update': ('managers', 'owners',),
-        }
-
-class MetaUserEntityType(UserEntityType):
-    meta = True
-
-
-class MetaRelationType(RelationType):
-    permissions = {
-        'read':   ('managers', 'users', 'guests',),
-        'add':    ('managers',),
-        'delete': ('managers',),
-        }
-    meta = True
-
-class UserRelationType(RelationType):
-    permissions = {
-        'read':   ('managers', 'users', 'guests',),
-        'add':    ('managers', 'users',),
-        'delete': ('managers', 'users',),
-        }
-
-class MetaUserRelationType(UserRelationType):
-    meta = True
-
-
-class AttributeRelationType(RelationType):
-    # just set permissions to None so default permissions are set
-    permissions = MARKER
-
-class MetaAttributeRelationType(AttributeRelationType):
-    meta = True
 
 # classes used to define relationships within entity type classes ##################
 
@@ -543,3 +526,58 @@ for basetype in BASE_TYPES:
     globals()[basetype] = type(basetype, (AbstractTypedAttribute,),
                                {'etype' : basetype})
 
+# provides a RichString class for convenience
+
+
+class RichString(String):
+    """Convenience RichString attribute type
+    The following declaration::
+
+      class Card(EntityType):
+          content = RichString(fulltextindexed=True, default_format='text/rest')
+
+    is equivalent to::
+
+      class Card(EntityType):
+          content_format = String(internationalizable=True,
+                                  default='text/rest', constraints=[format_constraint])
+          content  = String(fulltextindexed=True)
+    """
+    def __init__(self, default_format='text/plain', format_constraints=None, **kwargs):
+        self.default_format = default_format
+        self.format_constraints = format_constraints or [format_constraint]
+        super(RichString, self).__init__(**kwargs)
+
+
+# various derivated classes with some predefined values XXX deprecated
+
+class MetaEntityType(EntityType):
+    permissions = {
+        'read':   ('managers', 'users', 'guests',),
+        'add':    ('managers',),
+        'update': ('managers', 'owners',),
+        'delete': ('managers',),
+        }
+
+class MetaUserEntityType(EntityType):
+    pass
+
+class MetaRelationType(RelationType):
+    permissions = {
+        'read':   ('managers', 'users', 'guests',),
+        'add':    ('managers',),
+        'delete': ('managers',),
+        }
+
+class MetaUserRelationType(RelationType):
+    pass
+
+class MetaAttributeRelationType(RelationType):
+    # just set permissions to None so default permissions are set
+    permissions = MARKER
+
+
+__all__ += ('MetaEntityType', 'MetaUserEntityType',
+            'MetaRelationType', 'MetaUserRelationType',
+            'MetaAttributeRelationType')
+
