@@ -14,51 +14,14 @@ from decimal import Decimal
 from logilab.common.decorators import cached
 from logilab.common.compat import sorted
 from logilab.common.interface import implements
-from logilab.common.deprecation import deprecated_function
 
-from yams import BASE_TYPES, MARKER, ValidationError, BadSchemaDefinition
+import yams
+from yams import (BASE_TYPES, MARKER, ValidationError, BadSchemaDefinition,
+                  use_py_datetime)
 from yams.interfaces import (ISchema, IRelationSchema, IEntitySchema,
                              IVocabularyConstraint)
 from yams.constraints import BASE_CHECKERS, BASE_CONVERTERS, UniqueConstraint
 
-
-from mx.DateTime import today, now, DateTimeFrom, DateFrom, TimeFrom
-
-def use_py_datetime():
-    global DATE_FACTORY_MAP, KEYWORD_MAP
-
-    from datetime import datetime, date, time
-    from time import strptime as time_strptime
-
-    try:
-        strptime = datetime.strptime
-    except AttributeError: # py < 2.5
-        def strptime(value, format):
-            return datetime(*time_strptime(value, format)[:6])
-
-    def strptime_time(value, format='%H:%M'):
-        return time(*time_strptime(value, format)[3:6])
-
-    KEYWORD_MAP = {'Datetime.NOW' : datetime.now,
-                   'Datetime.TODAY': datetime.today,
-                   'Date.TODAY': date.today}
-    DATE_FACTORY_MAP = {
-        'Datetime' : lambda x: ':' in x and strptime(x, '%Y/%m/%d %H:%M') or strptime(x, '%Y/%m/%d'),
-        'Date' : lambda x : strptime(x, '%Y/%m/%d'),
-        'Time' : strptime_time
-        }
-
-try:
-    from mx.DateTime import today, now, DateTimeFrom, DateFrom, TimeFrom
-    KEYWORD_MAP = {'Datetime.NOW' : now,
-                   'Datetime.TODAY' : today,
-                   'Date.TODAY': today}
-    DATE_FACTORY_MAP = {'Datetime' : DateTimeFrom,
-                        'Date' : DateFrom,
-                        'Time' : TimeFrom}
-except ImportError:
-    use_py_datetime()
-
 def rehash(dictionary):
     """this function manually builds a copy of `dictionary` but forces
     hash values to be recomputed. Note that dict(d) or d.copy() don't
@@ -115,7 +78,6 @@ class ERSchema(object):
         assert erdef
         self.schema = schema
         self.type = erdef.name
-        self.meta = erdef.meta or False
         self.description = erdef.description or ''
         # mapping from action to groups
         try:
@@ -204,10 +166,11 @@ class ERSchema(object):
     def set_default_groups(self):
         """set default action -> groups mapping"""
         if self._groups:
-            # already initialized
-            pass
-            #assert not self.is_final(), \
-            #       'permission for final entities are not considered'
+            # already initialized, check everything is fine
+            for action, groups in self._groups.items():
+                assert isinstance(groups, tuple), \
+                       ('permission for action %s of %s isn\'t a tuple as '
+                        'expected' % (action, self.type))
         else:
             self._groups = self.get_default_groups()
 
@@ -283,16 +246,11 @@ class EntitySchema(ERSchema):
             # no permissions needed for final entities, access to them
             # is defined through relations
             return {'read': ('managers', 'users', 'guests',)}
-        elif self.meta:
-            return {'read': ('managers', 'users', 'guests',),
-                            'update': ('managers', 'owners',),
-                            'delete': ('managers',),
-                            'add': ('managers',)}
         else:
             return {'read': ('managers', 'users', 'guests',),
-                            'update': ('managers', 'owners',),
-                            'delete': ('managers', 'owners'),
-                            'add': ('managers', 'users',)}
+                    'update': ('managers', 'owners',),
+                    'delete': ('managers', 'owners'),
+                    'add': ('managers', 'users',)}
 
     # IEntitySchema interface #################################################
 
@@ -302,11 +260,6 @@ class EntitySchema(ERSchema):
         """
         return self.type in BASE_TYPES
 
-    def subjrproperty(self, rschema, prop):
-        return rschema.rproperty(self.type, rschema.objects(self.type)[0], prop)
-    def objrproperty(self, rschema, prop):
-        return rschema.rproperty(rschema.subjects(self.type)[0], self.type, prop)
-
     def is_subobject(self, strict=False):
         """return True if this entity type is contained by another. If strict,
         return True if this entity type *must* be contained by another.
@@ -399,8 +352,32 @@ class EntitySchema(ERSchema):
         assert len(objtypes) == 1
         return objtypes[0]
 
-    def rproperty(self, rtype, prop):
+    def subjrproperty(self, rtype, prop):
         """convenience method to access a property of a subject relation"""
+        return self.role_rproperty('subject', rtype, prop)
+
+    def objrproperty(self, rtype, prop):
+        """convenience method to access a property of an object relation"""
+        return self.role_rproperty('object', rtype, prop)
+
+    def role_rproperty(self, role, rtype, prop, ttype=None):
+        """convenience method to access a property of a relation according to
+        this schema role
+        """
+        if role == 'subject':
+            rschema = self.subject_relation(rtype)
+            if ttype is None:
+                ttype = rschema.objects(self)[0]
+            return rschema.rproperty(self, ttype, prop)
+        else:
+            assert role == 'object'
+            rschema = self.object_relation(rtype)
+            if ttype is None:
+                ttype = rschema.subjects(self)[0]
+            return rschema.rproperty(ttype, self, prop)
+
+    def rproperty(self, rtype, prop):
+        """convenience method to access a property of a final subject relation"""
         rschema = self.subject_relation(rtype)
         return rschema.rproperty(self, self.destination(rtype), prop)
 
@@ -475,7 +452,7 @@ class EntitySchema(ERSchema):
         attribute defined in the entity schema
         """
         for rschema, _ in self.attribute_definitions():
-            if not rschema.meta:
+            if not self.is_metadata(rschema):
                 return rschema
 
     def indexable_attributes(self):
@@ -539,9 +516,9 @@ class EntitySchema(ERSchema):
                     default = Decimal(default)
             elif attrtype in ('Date', 'Datetime', 'Time'):
                 try:
-                    default = KEYWORD_MAP['%s.%s' % (attrtype, default.upper())]()
+                    default = yams.KEYWORD_MAP['%s.%s' % (attrtype, default.upper())]()
                 except KeyError:
-                    default = DATE_FACTORY_MAP[attrtype](default)
+                    default = yams.DATE_FACTORY_MAP[attrtype](default)
             else:
                 default = unicode(default)
         return default
@@ -575,6 +552,15 @@ class EntitySchema(ERSchema):
                                  (rtype, self))
         return constraint.vocabulary()
 
+    def cardinality(self, rtype, role):
+        """return cardinality (a single letter in 1?*+) for the given relation"""
+        rschema = self.schema.rschema(rtype)
+        if role == 'subject':
+            targetschema = rschema.objects(self)[0]
+            return rschema.rproperty(self, targetschema, 'cardinality')[0]
+        targetschema = rschema.subjects(self)[0]
+        return rschema.rproperty(targetschema, self, 'cardinality')[1]
+
     def check(self, entity, creation=False, _=unicode):
         """check the entity and raises an ValidationError exception if it
         contains some invalid fields (ie some constraints failed)
@@ -735,6 +721,10 @@ class RelationSchema(ERSchema):
                 msg = "ambiguous relation %s: %s is final but not %s" % (
                     self.type, frschema, nfrschema)
                 raise BadSchemaDefinition(msg)
+        constraints = getattr(rdef, 'constraints', None)
+        if constraints:
+            for cstr in constraints:
+                cstr.check_consistency(subjschema, objschema, rdef)
         self.final = final
         # update our internal struct
         self._update(subjschema, objschema)
@@ -749,6 +739,9 @@ class RelationSchema(ERSchema):
                 pass
         else:
             self.init_rproperties(subjschema, objschema, rdef)
+        if self.inlined and self.rproperty(subjschema, objschema, 'cardinality')[0] in '*+':
+            raise BadSchemaDefinition('inlined relation %s can\'t have multiple cardinality for its subject %s'
+                                      % (self.type, subjschema))
         # update extremities schema
         subjschema.add_subject_relation(self)
         if self.symetric:
@@ -798,14 +791,10 @@ class RelationSchema(ERSchema):
             return {'read': ('managers', 'users', 'guests'),
                     'delete': ('managers', 'users', 'guests'),
                     'add': ('managers', 'users', 'guests')}
-        elif self.meta:
-            return {'read': ('managers', 'users', 'guests'),
-                            'delete': ('managers',),
-                            'add': ('managers',)}
         else:
             return {'read': ('managers', 'users', 'guests',),
-                            'delete': ('managers', 'users'),
-                            'add': ('managers', 'users',)}
+                    'delete': ('managers', 'users'),
+                    'add': ('managers', 'users',)}
 
     # relation definitions properties handling ################################
 
@@ -829,8 +818,6 @@ class RelationSchema(ERSchema):
         """return an iterator on (subject, object) of this relation"""
         return self._rproperties.iterkeys()
 
-    rproperty_keys = deprecated_function(iter_rdefs) # XXX bw compat
-
     def rdefs(self):
         """return a list of (subject, object) of this relation"""
         return self._rproperties.keys()
@@ -1152,6 +1139,8 @@ class Schema(object):
         try:
             return self._entities[etype]
         except KeyError:
+            if isinstance(etype, tuple):
+                etype = list(etype)
             raise KeyError('No entity named %s in schema' % etype)
 
     def relations(self):
